////////////////////////////////////////////////////////////////////////////////
//
// Date:       Feb. 27, 2007
// Module Name:    recoder_ctrl.v
// Description:    Implement a wishbone controller to control the recoder
//            function.  The module accepts Play/Stop and preset start
//            location.  Sample rate is adjusted by a parameter
//
// Revision:
// Revision 0.01 - File Created
//
// File format:    This file has been formated to use tabstop of 4
//
// Available Command:
//
// Development platform: Spartan-3E Starter kit
//
// Copyright (C ) 2007, Rick Huang
//   
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
// 
////////////////////////////////////////////////////////////////////////////////
//
// This state machine has been modified by Scott Gravenhorst.  (2011-10-29)
// Most of the wishbone has been removed because only the SDRAM is accessed.
// The wishbone SDRAM state machine remains almost unchanged.
//
// The new state machine is meant to be a general purpose interface to the SDRAM
// controller.  It works as follows:
//   
//   READ OPERATION:
//   Once the user detects USER_READY == 1: 
//     USER_RAM_ADDRESS is set
//     RW is set to 0
//     GO is asserted to 1
//   User waits for USER_READY == 0, then deasserts GO
//   User waits for USER_READY == 1
//   Read data is now available in USER_READ_DATA
//
//   WRITE OPERATION:
//   Once the user detects USER_READY == 1: 
//     USER_RAM_ADDRESS is set
//     USER_WRITE_DATA is set
//     RW is set to 1
//     GO is asserted to 1
//   User waits for USER_READY == 0, then deasserts GO
//   User waits for USER_READY == 1
//   Write operation is complete
//
////////////////////////////////////////////////////////////////////////////////

module SDRAM_ctrl_FSM
        (
        reset,      // BTN_S
        main_clk,                        // Main system clock 100 MHz 
        wADR_O, wSTB_O, wDAT_O, wDAT_I,
        wACK_I, wWE_O,
        RW,                                      // user Read or Write request
        USER_GO,                                 // user GO bit
        USER_WRITE_DATA,
        USER_READ_DATA,
        USER_RAM_ADDR,
        USER_READY,
        power_on,
        led
        );

  // Control input / UI output 
  input          reset;
  
  // Wishbone bus signal, bus master 
  input          main_clk;
  output  [31:0] wADR_O;         // ADDRESS DATA OUTPUT TO WISHBONE
  output  [31:0] wDAT_O;         // Wishbone WRITE DATA
  input   [31:0] wDAT_I;         // Wishbone READ DATA
  output         wSTB_O;
  output         wWE_O;          // 
  input          wACK_I;
  
  input          RW;                                      // user Read or Write request
  input          USER_GO;                                 // user GO bit
  input   [31:0] USER_WRITE_DATA;
  output  [31:0] USER_READ_DATA;
  input   [31:0] USER_RAM_ADDR;
  output         USER_READY;
  
  // connection to LEDs
  output  [7:0]  led;
  
  output         power_on;       // 0 = init in progress, 1 = good to go
  
//  reg [7:0] LED_reg = 0;
//  assign led = LED_reg;

  wire          RW;                                      // user Read or Write request
  wire          USER_GO;                                 // user GO bit
  wire   [31:0] USER_WRITE_DATA;
  reg    [31:0] USER_READ_DATA = 0;
  wire   [31:0] USER_RAM_ADDR;
  reg           USER_READY   = 0;

  // ** Main state machine ****************************************
  // * 
  // *

  reg    [31:0] dat_i_reg;             // register to store data read from wishbone input data bus
  reg    [4:0]  cmd_state;
  reg    [31:0] wADR_O;                // holds wishbone address data
  reg    [31:0] wDAT_O;                // holds wishbone write bus data
  reg    [3:0]  wbus_state;
  reg    [4:0]  wbus_timeout;
  reg           wSTB_O;
  reg           wWE_O;
  reg    [31:0] ram_addr;
  
  
  parameter  RAM_ADDR_MAX = 32'h24000000;
      
  parameter WBUS_IDLE     = 0,                // Wishbone bus state
            WBUS_RD_START = 1,
            WBUS_RD_ACKED = 2,
            WBUS_WR_START = 3,
            WBUS_WR_ACKED = 4,
            WBUS_ERR      = 5;
    
  parameter  WBUS_TIMEOUT_SET = 5'h1f;


  
  reg [14:0] power_on_cnt = 15'h4F00;      // a little more than 200 uSec  (202 uSec)
  reg        power_on     = 0;

  parameter CMD_IDLE          = 0,            // Command phase
            CMD_READ          = 1,
            CMD_READa         = 2,
            CMD_WRITE         = 3,
            CMD_WRITEa        = 4,
            CMD_STARTUP       = 5;

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// synchronizer

  reg GO = 0;
  reg GO_ = 0;
  
  always @ ( posedge main_clk )
    begin
    GO <= GO_;
    GO_ <= USER_GO;                  // sample USER_GO bit from other clock domain
    end

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Command State Machine

  always @ ( posedge main_clk ) 
    begin
    if ( (reset == 1) || (power_on_cnt != 0) ) 
      begin
      power_on_cnt <= power_on_cnt - 1; // power on counter causes this reset to remain set for a bit more than the 200 uSec required
      cmd_state    <= CMD_STARTUP;
      end 

    else

      begin
      case ( cmd_state )
      
        CMD_STARTUP:
          begin
          cmd_state  <= CMD_IDLE;
          USER_READY <= 1;
          power_on   <= 1;
          end

        CMD_IDLE: 
          begin
          ram_addr  <= (USER_RAM_ADDR | 32'h20000000);   // wishbone-ize the RAM address

          if ( GO )                            // User's GO signal
            begin
            USER_READY      <= 0;              // handshake signal to other clock domain
            if ( RW == 0 )                     // read / write command
              begin 
              cmd_state <= CMD_READ; 
              end
            else 
              begin 
              cmd_state <= CMD_WRITE;
              dat_i_reg <= USER_WRITE_DATA;
              end
            end
          else
            begin
            cmd_state <= CMD_IDLE; 
            end
          end

        CMD_READ:
          begin  
          if ( wbus_state == WBUS_IDLE )
            begin
            wADR_O          <= ram_addr;
            wbus_state      <= WBUS_RD_START;
            cmd_state       <= CMD_READa;
            end
          end

        CMD_READa:
          begin  
          if ( wbus_state == WBUS_IDLE )
            begin
            USER_READY      <= 1;
            cmd_state       <= CMD_IDLE;
            end
          end

        CMD_WRITE:
          begin  
          if ( wbus_state == WBUS_IDLE )
            begin
            wDAT_O          <= dat_i_reg;
            wADR_O          <= ram_addr;
            wbus_state      <= WBUS_WR_START;         // WRITE START COMMAND
            cmd_state       <= CMD_WRITEa;
            end
          end

        CMD_WRITEa:
          begin  
          if ( wbus_state == WBUS_IDLE )
            begin
            USER_READY      <= 1;                    // signal that the write is done
            cmd_state       <= CMD_IDLE;
            end
          end

        default: cmd_state <= CMD_IDLE;
      endcase

///////////////////////////////////////////////////////////////////////////////////////////////////    
///////////////////////////////////////////////////////////////////////////////////////////////////    
///////////////////////////////////////////////////////////////////////////////////////////////////    
// SDRAM state machine.  This FSM communicates with the DDR SDRAM IC.
    
      case ( wbus_state )            // WBus runs
        WBUS_RD_START:
          if ( wACK_I == 0 )
            begin
            wSTB_O          <= 1;          // set strobe to start a read cycle
            wWE_O           <= 0;          // set write enable off
            wbus_timeout    <= wbus_timeout - 1;
            if ( wbus_timeout == 0 ) 
              begin
              wbus_state <= WBUS_ERR;
              end
            end 
          else
            begin
            wbus_state <= WBUS_RD_ACKED;  // Received an ACK signal
            wSTB_O <= 0;                  // deassert strobe
            USER_READ_DATA <= wDAT_I;     // this is the data found on the wishbone input data bus
            end
      
        WBUS_RD_ACKED:
          begin
          if ( wACK_I == 0 ) wbus_state <= WBUS_IDLE;
          end
      
        WBUS_WR_START:
          begin
          if ( wACK_I == 0 )
            begin
            wSTB_O <= 1;                  // set strobe to start a write cycle
            wWE_O  <= 1;                  // SET WRITE ENABLE ON
            wbus_timeout <= wbus_timeout - 1;
            if ( wbus_timeout == 0 ) wbus_state <= WBUS_ERR;
            end 
          else
            begin
            wbus_state <= WBUS_WR_ACKED;  // Received a ACK signal
            wSTB_O <= 0;                  // deassert strobe
            wWE_O  <= 0;                  // CLEAR WRITE ENABLE
            end
          end
      
        WBUS_WR_ACKED:
          begin
          if ( wACK_I == 0 ) wbus_state <= WBUS_IDLE;
          end
          
        WBUS_IDLE:
          begin
          wSTB_O <= 0;
          wbus_timeout <= WBUS_TIMEOUT_SET;
          end
          
        WBUS_ERR:
          begin
          wSTB_O <= 0;
          wbus_state <= WBUS_IDLE;
          end
        endcase
      end
    end

endmodule
